from collections import OrderedDict, deque
from typing import Callable, Dict

import streamlit as st


def render_chat_log_message(msg: Dict, markdown=True):
    msg = msg.copy()
    render = st.markdown if markdown else st.text
    if msg.get("user_role"):
        messages = msg.pop('messages')
        with st.chat_message(name="user", avatar='🧑'):
            for msg_item in messages:
                # https://github.com/streamlit/streamlit/issues/7978
                render(f"`{msg_item['role']}`: {msg_item['content']}")
            st.write(msg)
    elif msg.get("assistant_role"):
        with st.chat_message(name="assistant", avatar='🤖'):
            ass_content = msg.pop('assistant', None)
            render(ass_content)
            st.write(msg)
    else:
        print(f"{msg=}")


class ChatData:
    def __init__(self, max_len: int):
        self.data = OrderedDict()
        self.max_len = max_len

    def add_message(self, message, **callback_kwargs):
        uid = message.pop('uid')
        if message.get("user_role"):
            self.data.setdefault(uid, {'user_role': message})
        else:
            msg_item = self.data.get(uid, {})
            msg_item['assistant_role'] = message
        if len(self.data) >= self.max_len:
            self.data.popitem(last=False)
        render_chat_log_message(message, **callback_kwargs)

    def render_all_messages(self, **callback_kwargs):
        for uid, msg in self.data.items():
            message = msg.get("user_role")
            if message:
                render_chat_log_message(message, **callback_kwargs)
            message = msg.get("assistant_role")
            if message:
                render_chat_log_message(message, **callback_kwargs)



from __future__ import annotations

import time
from itertools import cycle
from typing import List, Literal, Optional, Union

import attrs
import orjson
from fastapi import Request
from fastapi.responses import Response, StreamingResponse

from ...decorators import (
    async_random_sleep,
    async_token_rate_limit_auth_level,
    random_sleep,
)
from ...helper import get_unique_id
from ...settings import FWD_KEY, token_interval_conf
from .tokenizer import TIKTOKEN_VALID, count_tokens, encode_as_pieces


@attrs.define(slots=True)
class ChatMessage:
    role: Literal["user", "assistant", "system"]
    content: Optional[str] = None


@attrs.define(slots=True)
class ChatCompletionMessage:
    role: Literal["user", "assistant", "system"]
    content: Optional[str] = None
    tool_calls: Optional[list] = None


@attrs.define(slots=True)
class Function:
    name: str
    arguments: dict


@attrs.define(slots=True)
class ChoiceToolCall:
    id: str
    function: Function
    type: str = 'function'


@attrs.define(slots=True)
class DeltaFunction:
    name: str
    arguments: str


@attrs.define(slots=True)
class ChoiceDeltaToolCall:
    index: int = 0
    id: Optional[str] = None
    type: Optional[str] = None
    function: Optional[dict] = None


@attrs.define(slots=True)
class DeltaMessage:
    role: Optional[Literal["user", "assistant", "system"]] = None
    content: Optional[str] = None
    tool_calls: Optional[ChoiceDeltaToolCall] = None


@attrs.define(slots=True)
class ChatCompletionRequest:
    model: str
    messages: List[ChatMessage]
    temperature: Optional[float] = None
    top_p: Optional[float] = None
    max_length: Optional[int] = None
    stream: Optional[bool] = False


@attrs.define(slots=True)
class ChatCompletionChoice:
    index: int
    message: ChatCompletionMessage
    finish_reason: str


@attrs.define(slots=True)
class ChatCompletionResponseStreamChoice:
    index: int
    delta: DeltaMessage
    finish_reason: Optional[Literal["stop", "length"]]


@attrs.define(slots=True)
class ChatCompletionsResponse:
    id: str
    object: Literal["chat.completion", "chat.completion.chunk"]
    created: Optional[int]
    model: Optional[str]
    choices: List[Union[ChatCompletionChoice, ChatCompletionResponseStreamChoice]]
    usage: Optional[dict] = None
    system_fingerprint: str = "fp_0123456789"


corpus = [
    """他还太年轻，尚不知道回忆总是会抹去坏的，夸大好的，也正是由于这种玄妙，我们才得以承担过去的重负。
"He was still too young to know that the heart's memory eliminates the bad and magnifies the good, and that thanks to this artifice we manage to endure the burden of the past."
""",
    """当他年轻的时候，他认为时间是个不会耗尽的海洋。现在他已经成了一个老人，而他每天都更加清楚，时间是个窄窄的峡谷。
"When he was young, he thought time was a boundless ocean; now as an old man, he knew more each day that time was a narrow strait."
""",
    """我们的身体认识彼此，就像两本同样的书：书里的内容，书的纸张，书的形状和包装，我都爱得不能自拔。
"Our bodies knew each other as if they were two volumes of the same book: the text, the paper, the shape and cover of the book, I loved them all without reserve."
""",
    """多少人爱你青春欢畅的时辰，爱慕你的美丽，假意或者真心，只有一个人爱你那朝圣者的灵魂，爱你衰老了的脸上痛苦的皱纹
"Many people love you during your youthful and cheerful moments, admiring your beauty, either pretentiously or sincerely. Only one person loves the soul of a pilgrim within you, and the painful wrinkles on your aged face."
""",
    """他离她那么近，甚至能听到她每一次的呼吸，闻到她身上散发的馨香，在此生余下的岁月中，他正是靠着这种馨香来辨认她。
"He is so close to her that he can even hear each of her breaths and smell the fragrance that emanates from her. For the remaining years of his life, it is this fragrance by which he identifies her."
 """,
]

sentences = cycle(corpus)


@async_token_rate_limit_auth_level(token_interval_conf, FWD_KEY)
async def stream_generate(
    model: str, content: str | None, tool_calls: list | None, request: Request
):
    created = int(time.time())
    id = f"chatcmpl-{get_unique_id()}"

    if tool_calls:
        function_name = tool_calls[0]['function']['name']
        function_arguments = tool_calls[0]['function']['arguments']
        texts = encode_as_pieces(function_arguments)
    else:
        texts = encode_as_pieces(content)

    delta = DeltaMessage()

    choice_data = ChatCompletionResponseStreamChoice(
        index=0, delta=delta, finish_reason=None
    )
    chunk = ChatCompletionsResponse(
        id=id,
        model=model,
        choices=[choice_data],
        object="chat.completion.chunk",
        created=created,
        system_fingerprint="fp_0123456789",
    )

    def serialize_delta(
        role=None, content=None, delta_tool_calls=None, finish_reason=None
    ):
        if role:
            delta.role = role
        if content:
            delta.content = content
        if tool_calls:
            delta.tool_calls = delta_tool_calls

        choice_data.finish_reason = finish_reason
        choice_data.delta = delta

        chunk.choices = [choice_data]

        return (
            b'data: '
            + orjson.dumps(
                attrs.asdict(chunk, filter=attrs.filters.exclude(type(None)))
            )
            + b'\n\n'
        )

    if tool_calls:
        yield serialize_delta(
            role="assistant",
            delta_tool_calls=[
                {
                    'index': 0,
                    'id': f"call_{get_unique_id()}",
                    'function': {"name": function_name, "arguments": ""},
                    'type': 'function',
                }
            ],
        )
    else:
        yield serialize_delta(role="assistant", content="")

    delta = DeltaMessage()
    for content in texts:
        if tool_calls:
            yield serialize_delta(
                delta_tool_calls=[
                    {
                        'index': 0,
                        'id': None,
                        'function': {"name": None, "arguments": content},
                        'type': None,
                    }
                ],
            )
        else:
            yield serialize_delta(content=content)

    delta = DeltaMessage()
    yield serialize_delta(finish_reason="tool_calls" if tool_calls else "stop")

    yield b'data: [DONE]\n\n'


@async_token_rate_limit_auth_level(token_interval_conf, FWD_KEY)
async def stream_generate_efficient(
    model: str, content: str | None, tool_calls: list | None, request: Request
):
    """More efficient (use dict) version of stream_generate
    Args:
        model (str): The model to use.
        content (str): content.
        tool_calls (list | None): tool_calls list.
        request (Request): A FastAPI request object. For rate limit.
    """
    created = int(time.time())
    id = f"chatcmpl-{get_unique_id()}"
    if tool_calls:
        function_name = tool_calls[0]['function']['name']
        function_arguments = tool_calls[0]['function']['arguments']
        texts = encode_as_pieces(function_arguments)
    else:
        texts = encode_as_pieces(content)

    delta = {}
    choice_data = {"index": 0, "delta": delta, "finish_reason": None}
    chunk = {
        "id": id,
        "model": model,
        "choices": [choice_data],
        "object": "chat.completion.chunk",
        "created": created,
        "system_fingerprint": "fp_0123456789",
    }

    def serialize_delta(
        role=None, content=None, delta_tool_calls=None, finish_reason=None
    ):
        if role:
            delta['role'] = role
        if content:
            delta['content'] = content
        if delta_tool_calls:
            delta['tool_calls'] = delta_tool_calls
            delta['content'] = content

        choice_data['finish_reason'] = finish_reason
        choice_data['delta'] = delta

        chunk['choices'] = [choice_data]

        return b'data: ' + orjson.dumps(chunk) + b'\n\n'

    if tool_calls:
        yield serialize_delta(
            role="assistant",
            delta_tool_calls=[
                {
                    'index': 0,
                    'id': f"call_{get_unique_id()}",
                    'function': {"name": function_name, "arguments": ""},
                    'type': 'function',
                }
            ],
        )
    else:
        yield serialize_delta(role="assistant", content="")

    delta = {}
    for content in texts:
        if tool_calls:
            yield serialize_delta(
                delta_tool_calls=[
                    {
                        'index': 0,
                        'id': None,
                        'function': {"name": None, "arguments": content},
                        'type': None,
                    }
                ],
            )
        else:
            yield serialize_delta(content=content)

    delta = {}
    yield serialize_delta(finish_reason="tool_calls" if tool_calls else "stop")

    yield b'data: [DONE]\n\n'


@random_sleep(min_time=1, max_time=2)
def generate(model: str, content: str | None, tool_calls: list | None, usage: dict):
    created = int(time.time())
    id = f"chatcmpl-{get_unique_id()}"

    choice_data = ChatCompletionChoice(
        index=0,
        message=ChatCompletionMessage(
            role="assistant", content=content, tool_calls=tool_calls
        ),
        finish_reason="tool_calls" if tool_calls else "stop",
    )

    data = ChatCompletionsResponse(
        id=id,
        created=created,
        model=model,
        choices=[choice_data],
        object="chat.completion",
        usage=usage,
    )

    return orjson.dumps(
        attrs.asdict(data),
        option=orjson.OPT_APPEND_NEWLINE,  # not necessary
    )


@attrs.define(slots=True)
class ModelInferResult:
    content: str
    usage: dict


def model_inference(model: str, messages: List):
    sentence = next(sentences)

    if TIKTOKEN_VALID:
        usage = count_tokens(messages, sentence, 'gpt-3.5-turbo')
    else:
        usage = {"prompt_tokens": 0, "completion_tokens": 0, "total_tokens": 0}

    return ModelInferResult(content=sentence, usage=usage)


async def chat_completions_benchmark(request: Request):
    payload = await request.json()
    model = payload.get("model", 'robot')
    stream = payload.get("stream", False)
    messages = payload.get("messages", [])

    model_result = model_inference(model, messages)

    if stream:
        return StreamingResponse(
            stream_generate_efficient(model, model_result.content, None, request),
            media_type="text/event-stream",
        )
    else:
        return Response(
            content=generate(model, model_result.content, None, model_result.usage),
            media_type="application/json",
        )



import json
import os
from typing import Dict, List, Literal, Optional, Tuple, Union

from attrs import asdict, define, field, filters

from ..settings import *


class Base:
    def to_dict(self, drop_none=True):
        if drop_none:
            return asdict(self, filter=filters.exclude(type(None)))
        return asdict(self)

    def to_dict_str(self):
        return {k: str(v) for k, v in self.to_dict(drop_none=True).items()}


@define(slots=True)
class ForwardItem(Base):
    base_url: str
    route: str
    type: Literal["openai", "general"] = field(default="general")


@define(slots=True)
class Forward(Base):
    forward: List[ForwardItem] = [
        ForwardItem(base_url="https://api.openai.com", route="/", type="openai"),
        ForwardItem(
            base_url="https://generativelanguage.googleapis.com",
            route="/gemini",
            type="general",
        ),
    ]

    def convert_to_env(self, set_env=False):
        env_dict = {'FORWARD_CONFIG': json.dumps([i.to_dict() for i in self.forward])}

        if set_env:
            os.environ.update(env_dict)
        return env_dict


@define(slots=True)
class CacheConfig(Base):
    backend: str = 'LevelDB'
    root_path_or_url: str = './FLAXKV_DB'
    default_request_caching_value: bool = True
    cache_openai: bool = False
    cache_general: bool = False
    cache_routes: List = ['/v1/chat/completions']

    def convert_to_env(self, set_env=False):

        env_dict = {}

        env_dict['CACHE_OPENAI'] = str(self.cache_openai)
        env_dict['CACHE_GENERAL'] = str(self.cache_general)

        env_dict['CACHE_BACKEND'] = self.backend
        env_dict['CACHE_ROOT_PATH_OR_URL'] = self.root_path_or_url
        env_dict['DEFAULT_REQUEST_CACHING_VALUE'] = str(
            self.default_request_caching_value
        )
        env_dict['CACHE_ROUTES'] = json.dumps(self.cache_routes)

        if set_env:
            os.environ.update(env_dict)
        return env_dict


@define(slots=True)
class RateLimitType(Base):
    route: str
    value: List[Dict[str, str]]


@define(slots=True)
class RateLimit(Base):
    backend: str = ''
    global_rate_limit: str = 'inf'
    token_rate_limit: List[RateLimitType] = [
        RateLimitType(
            route="/v1/chat/completions",
            value=[{"level": '0', "rate_limit": "60/second"}],
        ),
        RateLimitType(
            route="/v1/completions", value=[{"level": '0', "rate_limit": "60/second"}]
        ),
    ]
    req_rate_limit: List[RateLimitType] = [
        RateLimitType(
            route="/v1/chat/completions",
            value=[{"level": '0', "rate_limit": "100/2minutes"}],
        ),
        RateLimitType(
            route="/v1/completions", value=[{"level": '0', "rate_limit": "60/minute"}]
        ),
        RateLimitType(
            route="/v1/embeddings", value=[{"level": '0', "rate_limit": "100/2minutes"}]
        ),
    ]
    iter_chunk: Literal['one-by-one', 'efficiency'] = 'one-by-one'
    strategy: Literal[
        'fixed_window', 'moving-window', 'fixed-window-elastic-expiry'
    ] = 'moving-window'

    def convert_to_env(self, set_env=False):
        env_dict = {}
        env_dict['GLOBAL_RATE_LIMIT'] = self.global_rate_limit
        env_dict['RATE_LIMIT_STRATEGY'] = self.strategy
        env_dict['TOKEN_RATE_LIMIT'] = json.dumps(
            {i.route: i.value for i in self.token_rate_limit if i.route and i.value}
        )
        env_dict['REQ_RATE_LIMIT'] = json.dumps(
            {i.route: i.value for i in self.req_rate_limit if i.route and i.value}
        )
        env_dict['ITER_CHUNK_TYPE'] = self.iter_chunk
        if set_env:
            os.environ.update(env_dict)
        return env_dict


@define(slots=True)
class ApiKey(Base):
    openai_key: Dict = {"": "0"}
    forward_key: Dict = {"": 0}
    level: Dict = {1: ["gpt-3.5-turbo"]}

    def convert_to_env(self, set_env=False):
        env_dict = {}
        openai_key_dict = {}
        for key, value in self.openai_key.items():
            value: str
            values = value.strip().replace('，', ',').split(',')
            openai_key_dict[key] = [int(i) for i in values]
        env_dict['OPENAI_API_KEY'] = json.dumps(openai_key_dict)
        env_dict['FORWARD_KEY'] = json.dumps(self.forward_key)
        env_dict['LEVEL_MODELS'] = json.dumps(self.level)
        if set_env:
            os.environ.update(env_dict)
        return env_dict


@define(slots=True)
class Log(Base):
    LOG_GENERAL: bool = True
    LOG_OPENAI: bool = True

    def convert_to_env(self, set_env=False):
        env_dict = {}
        env_dict['LOG_GENERAL'] = str(self.LOG_GENERAL)
        env_dict['LOG_OPENAI'] = str(self.LOG_OPENAI)
        if set_env:
            os.environ.update(env_dict)
        return env_dict


@define(slots=True)
class Config(Base):
    forward: Forward = Forward()

    api_key: ApiKey = ApiKey()

    cache: CacheConfig = CacheConfig()

    rate_limit: RateLimit = RateLimit()

    log: Log = Log()

    timezone: str = 'Asia/Shanghai'
    timeout: int = 6
    benchmark_mode: bool = False
    proxy: str = ''
    webui_restart_port: int = 15555
    webui_log_port: int = 15556
    default_stream_response: bool = True

    def convert_to_env(self, set_env=False):
        env_dict = {}
        env_dict.update(self.forward.convert_to_env())
        env_dict.update(self.api_key.convert_to_env())
        env_dict.update(self.cache.convert_to_env())
        env_dict.update(self.rate_limit.convert_to_env())
        env_dict.update(self.log.convert_to_env())

        env_dict['TZ'] = self.timezone
        env_dict['TIMEOUT'] = str(self.timeout)
        env_dict['BENCHMARK_MODE'] = str(self.benchmark_mode)
        env_dict['WEBUI_RESTART_PORT'] = str(self.webui_restart_port)
        env_dict['WEBUI_LOG_PORT'] = str(self.webui_log_port)
        env_dict['DEFAULT_STREAM_RESPONSE'] = str(self.default_stream_response)

        if self.proxy:
            env_dict['PROXY'] = self.proxy

        if set_env:
            os.environ.update(env_dict)
        return env_dict

    def come_from_env(self):

        self.timeout = TIMEOUT
        self.timezone = os.environ.get('TZ', 'Asia/Shanghai')
        self.benchmark_mode = BENCHMARK_MODE
        self.proxy = PROXY or ""
        self.log.LOG_OPENAI = LOG_OPENAI
        self.log.LOG_GENERAL = LOG_GENERAL

        self.rate_limit.strategy = RATE_LIMIT_STRATEGY
        self.rate_limit.global_rate_limit = GLOBAL_RATE_LIMIT
        self.rate_limit.iter_chunk = ITER_CHUNK_TYPE
        self.rate_limit.backend = RATE_LIMIT_BACKEND or self.rate_limit.backend
        self.rate_limit.req_rate_limit = [
            RateLimitType(key, value) for key, value in req_rate_limit_dict.items()
        ] or self.rate_limit.req_rate_limit
        self.rate_limit.token_rate_limit = [
            RateLimitType(key, value) for key, value in token_rate_limit_conf.items()
        ] or self.rate_limit.token_rate_limit

        self.cache.backend = CACHE_BACKEND
        self.cache.root_path_or_url = CACHE_ROOT_PATH_OR_URL
        self.cache.default_request_caching_value = DEFAULT_REQUEST_CACHING_VALUE
        self.cache.cache_openai = CACHE_OPENAI or self.cache.cache_openai
        self.cache.cache_general = CACHE_GENERAL or self.cache.cache_general
        self.cache.cache_routes = list(CACHE_ROUTE_SET) or self.cache.cache_routes
        self.api_key.level = LEVEL_MODELS or self.api_key.level
        self.api_key.openai_key = {
            key: ','.join([str(i) for i in value])
            for key, value in OPENAI_API_KEY.items()
        } or self.api_key.openai_key
        self.api_key.forward_key = FWD_KEY or self.api_key.forward_key
        self.forward.forward = [ForwardItem(**i) for i in FORWARD_CONFIG]

        return self



try:
    import tiktoken

    TIKTOKEN_VALID = True

except ImportError:
    print("Warning: `tiktoken` not found. Install with `pip install tiktoken`.")
    TIKTOKEN_VALID = False


def encode_as_pieces(sentence):
    words = []

    delimiters = {'：', '。', '，', '、', '！', '？', ':', '.', ',', '!', '?'}

    buffer = ""
    for char in sentence:
        if '\u4e00' <= char <= '\u9fff' or char in delimiters:
            if buffer:
                words.append(buffer)
                buffer = ""
            words.append(char)
        elif char == ' ':
            if buffer:
                words.append(buffer)
                buffer = ""
            words.append(char)
        else:
            buffer += char

    if buffer:
        words.append(buffer)

    return words


def count_tokens(messages, assistant_content, model="gpt-3.5-turbo"):
    """Return the usage information of tokens in the messages list."""
    # https://github.com/openai/openai-cookbook/blob/main/examples/How_to_format_inputs_to_ChatGPT_models.ipynb
    if not TIKTOKEN_VALID:
        print("Warning: `tiktoken` not found. Install with `pip install tiktoken`.")
        raise ImportError
    try:
        encoding = tiktoken.encoding_for_model(model)
    except KeyError:
        print("Warning: model not found. Using cl100k_base encoding.")
        encoding = tiktoken.get_encoding("cl100k_base")

    count_name = 0
    messages_len = len(messages)
    content = ""
    for i in messages:
        content += i['content'] + i['role']
        name = i.get('name')
        if name:
            count_name += 1

    tokens_per_message = 3
    tokens_per_name = 1

    history_tokens = (
        len(encoding.encode(content))
        + (messages_len + 1) * tokens_per_message
        + count_name * tokens_per_name
    )

    assis_tokens = len(encoding.encode(assistant_content))

    total_tokens = history_tokens + assis_tokens
    return {
        "prompt_tokens": history_tokens,
        "completion_tokens": assis_tokens,
        "total_tokens": total_tokens,
    }



from flaxkv.pack import decode, encode
from loguru import logger

from ..settings import (
    CACHE_OPENAI,
    CACHE_ROUTE_SET,
    CHAT_COMPLETION_ROUTE,
    EMBEDDING_ROUTE,
)
from .chat.response import gen_response, get_cached_chat_response
from .database import db_dict
from .embedding.response import get_cached_embedding_response


def get_cached_response(
    payload, payload_info, valid_payload, route_path, request, **kwargs
):
    if route_path == EMBEDDING_ROUTE:
        # todo: use general cache replace may be better?
        return get_cached_embedding_response(
            payload_info, valid_payload, request, **kwargs
        )
    elif route_path == CHAT_COMPLETION_ROUTE:
        return get_cached_chat_response(payload_info, valid_payload, request, **kwargs)

    elif route_path in CACHE_ROUTE_SET:
        return get_cached_generic_response(payload, request, route_path)
    else:
        return None, None


def cache_response(cache_key, target_info, route_path, chunk_list):
    if (
        target_info
        and CACHE_OPENAI
        and route_path == CHAT_COMPLETION_ROUTE
        and cache_key is not None
    ):
        cached_value = db_dict.get(cache_key, {"data": []})["data"]
        if len(cached_value) < 10:
            cached_value.append(target_info["assistant"])
            db_dict[cache_key] = {
                "data": cached_value,
                "route_path": route_path,
            }
    elif (
        target_info
        and CACHE_OPENAI
        and route_path == EMBEDDING_ROUTE
        and cache_key is not None
    ):
        cached_value = bytes(target_info["buffer"])
        db_dict[cache_key] = {
            "data": cached_value,
            "route_path": route_path,
        }
    else:
        cache_generic_response(cache_key, route_path, chunk_list)


def cache_generic_response(cache_key, buffer_list, route_path, max_cache=3):
    if cache_key and route_path in CACHE_ROUTE_SET:
        value_list = db_dict.get(cache_key, [])

        if len(value_list) < max_cache:
            value_list.append({"data": buffer_list})
            db_dict[cache_key] = value_list


def get_cached_generic_response(payload: bytes, request, route_path):

    if route_path not in CACHE_ROUTE_SET:
        return None, None

    try:
        cache_key = encode(payload)
        value_list = db_dict.get(cache_key)
        if value_list is None:
            return None, cache_key

        idx = -1
        buffer_list = value_list[idx]["data"]
        if len(buffer_list) == 1:
            logger.info(f'generic cache >>>>>>>> [cache hit]')
            logger.debug(f"cache result: {buffer_list[0]}")
        elif len(buffer_list) > 1:
            logger.info(f'generic cache >>>>>>>> [stream cache hit]')
        else:
            raise ValueError(f"Invalid buffer list: {buffer_list}")
        return gen_response(buffer_list, request), cache_key

    except:
        return None, None



from __future__ import annotations

import random
from typing import List

from fastapi.responses import Response, StreamingResponse
from flaxkv.pack import encode
from loguru import logger

from ...settings import CACHE_OPENAI, FWD_KEY
from ..database import db_dict
from .chat_completions import (
    async_token_rate_limit_auth_level,
    generate,
    stream_generate_efficient,
    token_interval_conf,
)


def construct_cache_key(payload_info: dict):
    elements = [
        payload_info["n"],
        payload_info['messages'],
        payload_info['model'],
        payload_info["max_tokens"],
        payload_info['response_format'],
        payload_info['seed'],
        # payload_info['temperature'],
        payload_info["tools"],
        payload_info["tool_choice"],
    ]

    return encode(elements)


def get_response_from_key(key, payload_info, request, **kwargs):
    value = db_dict[key]
    cache_values = value['data']
    idx = random.randint(0, len(cache_values) - 1) if len(cache_values) > 1 else 0
    uid = payload_info["uid"]
    logger.info(f'chat uid: {uid} >>>{idx}>>>> [cache hit]')
    # todo: handle multiple choices
    cache_value = cache_values[idx]
    if isinstance(cache_value, list):
        text = None
        tool_calls = cache_value
        is_tool_calls = True
    else:
        text = cache_value
        tool_calls = None
        is_tool_calls = False

    logger_instance = kwargs.get("logger_instance")
    if logger_instance:
        result_info = {
            "assistant": cache_value,
            "is_tool_calls": is_tool_calls,
            'uid': uid,
        }
        if logger_instance.webui:
            logger_instance.q.put({"uid": uid, "result": result_info})

        logger_instance.log_result(result_info)

    if payload_info["stream"]:
        return StreamingResponse(
            stream_generate_efficient(
                payload_info['model'],
                text,
                tool_calls,
                request,
            ),
            status_code=200,
            media_type="text/event-stream",
        )

    else:
        usage = {
            "prompt_tokens": 0,
            "completion_tokens": 0,
            "total_tokens": 0,
        }
        return Response(
            content=generate(payload_info['model'], text, tool_calls, usage),
            media_type="application/json",
        )


def get_cached_chat_response(payload_info, valid_payload, request, **kwargs):
    """
    Attempts to retrieve a cached response based on the current request's payload information.

    This function constructs a cache key based on various aspects of the request payload,
    checks if the response for this key has been cached, and if so, constructs and returns
    the appropriate cached response.

    Returns:
        Tuple[Union[Response, None], Union[str, None]]:
            - Response (Union[Response, None]): The cached response if available; otherwise, None.
            - cache_key (Union[str, None]): The constructed cache key for the request. None if caching is not applicable.

    Note:
        If a cache hit occurs, the cached response is immediately returned without contacting the external server.
    """
    if not (CACHE_OPENAI and valid_payload):
        return None, None

    cache_key = construct_cache_key(payload_info)

    if payload_info['caching'] and cache_key in db_dict:
        return (
            get_response_from_key(cache_key, payload_info, request, **kwargs),
            cache_key,
        )

    return None, cache_key


@async_token_rate_limit_auth_level(token_interval_conf, FWD_KEY)
async def stream_generate(buffer_list: List, request):
    for buffer in buffer_list:
        yield buffer


def gen_response(buffer_list, request):
    if len(buffer_list) > 1:
        return StreamingResponse(
            stream_generate(buffer_list, request),
            status_code=200,
            media_type="application/octet-stream",
        )
    else:
        return Response(
            buffer_list[0],
            status_code=200,
            media_type="application/octet-stream",
        )



import ast
import pickle
import threading

import orjson
import pandas as pd
import streamlit as st
import zmq
from flaxkv.helper import SimpleQueue

from openai_forward.config.interface import *
from openai_forward.webui.chat import ChatData
from openai_forward.webui.helper import mermaid

st.set_page_config(
    page_title="Openai Forward ",
    page_icon="🚀",
    # layout="wide",
    initial_sidebar_state="expanded",
    menu_items={
        'Get help': 'https://github.com/KenyonY/openai-forward',
        'Report a bug': "https://github.com/KenyonY/openai-forward/issues",
        'About': "# Openai Forward",
    },
)


@st.cache_resource
def get_global_vars():
    context = zmq.Context()
    socket = context.socket(zmq.REQ)
    # socket.setsockopt(zmq.CONNECT_TIMEOUT, 20000) # 20s
    openai_forward_host = os.environ.get("OPENAI_FORWARD_HOST", "localhost")
    restart_port = int(os.environ.get('WEBUI_RESTART_PORT', 15555))
    socket.connect(f"tcp://{openai_forward_host}:{restart_port}")

    log_socket = context.socket(zmq.DEALER)
    webui_log_port = int(os.environ.get("WEBUI_LOG_PORT", 15556))
    log_socket.connect(f"tcp://{openai_forward_host}:{webui_log_port}")
    log_socket.send_multipart([b"/subscribe", b"0"])

    def worker(log_socket: zmq.Socket, q: SimpleQueue):
        while True:
            message = log_socket.recv_multipart()
            uid, msg = message
            q.put((uid, msg))

    q = SimpleQueue(maxsize=100)
    threading.Thread(target=worker, args=(log_socket, q)).start()
    config = Config().come_from_env()
    chat_data = ChatData(100)
    return {
        "socket": socket,
        "log_socket": log_socket,
        "q": q,
        "config": config,
        "chat_data": chat_data,
    }


if 'socket' not in st.session_state:
    st.session_state.update(get_global_vars())

config = st.session_state['config']

with st.sidebar:
    selected_section = st.radio(
        "Select a configuration section",
        (
            "Forward",
            "API Key && Level",
            "Cache",
            "Rate Limit",
            "Other",
            "Real-time Logs",
            "Playground",
            "Statistics",
        ),
    )

    if st.button(
        "Apply and Restart", help="Saving configuration and reloading openai forward"
    ):
        with st.spinner("Saving configuration and reloading openai forward..."):
            env_dict = config.convert_to_env(set_env=False)
            socket = st.session_state['socket']
            socket.send(pickle.dumps(env_dict))
            message: bytes = socket.recv()

        st.success(message.decode())

    def generate_env_content():
        env_dict = config.convert_to_env(set_env=False)
        env_content = "\n".join([f"{key}={value}" for key, value in env_dict.items()])
        return env_content

    if st.button(
        "Export to .env file",
    ):
        # Deferred data for download button: https://github.com/streamlit/streamlit/issues/5053
        download = st.download_button(
            use_container_width=True,
            label="Export",
            data=generate_env_content(),
            file_name="config.env",
            mime="text/plain",
        )


def display_forward_configuration():
    forward_config = config.forward

    st.subheader("AI Forward")
    with st.form("forward_configuration", border=False):
        df = pd.DataFrame([i.to_dict() for i in forward_config.forward])
        edited_df = st.data_editor(
            df, num_rows="dynamic", key="editor1", use_container_width=True
        )

        with st.expander("See explanation"):
            df_demo = pd.DataFrame(
                [
                    {
                        'base_url': 'https://api.openai.com',
                        'route': '/',
                        'type': 'openai',
                    },
                    {
                        "base_url": "https://generativelanguage.googleapis.com",
                        "route": "/gemini",
                        "type": "general",
                    },
                ]
            )

            st.write(df_demo)
            st.write(
                "> 在以上设置下:  \n"
                "> - openai转发地址为： http://localhost:8000/  \n"
                "> - type=openai转发下的服务需要满足openai api 格式才能被正确解析  \n\n"
                "> - gemini转发地址为： http://localhost:8000/gemini  \n"
                "> - type=general转发下的服务可以是任何服务（暂不支持websocket)"
            )

        st.write("#")

        submitted = st.form_submit_button("Save", use_container_width=True)
        if submitted:
            forward_config.forward = [
                ForwardItem(row["base_url"], row["route"], row["type"])
                for i, row in edited_df.iterrows()
                if row["route"] is not None and row["base_url"] is not None
            ]

            print(forward_config.convert_to_env())


def display_api_key_configuration():
    with st.expander("See explanation"):
        st.write(
            """
> - 不同level对应不同权限，可以定义不同权限可以访问哪些模型。 例如, level为1，只可访问gpt-3.5-turbo；level为2，可访问embedding 和tts等等。(level为0，全权限，可访问任何模型，不可更改)
> - 对于api key而言，level的意义是区分该api key自身是否有权限访问哪些模型。每个api key可对应多个level，因为这些level不必是从属关系。
> - 对于forward key而言，每个key对应一个level，level表示它可以访问该level对应的所有api key。
"""
        )
        mermaid(
            """
graph TD
fk1(FK1) --> level0(Level 0)
fk2(FK2) --> level0
level0 --> sk1(SK1)
level0 --> sk2(SK2)
level0 --> sk3(SK3)

fk3(FK3) --> level1(Level 1)
level1 --> sk4(SK4)
level1 --> sk3(SK3)

fk(FK_x) --> level_n(Level n)
level_n --> sk_n(SK_n)
"""
        )

    # check openai models:
    # from openai import OpenAI
    # from rich import print
    # client = OpenAI(api_key="sk-")
    # openai_model_list = [i.id for i in client.models.list()]
    # openai_model_list.sort()
    # print(openai_model_list)
    # print(len(openai_model_list))

    openai_model_list = [
        'babbage-002',
        'dall-e-2',
        'dall-e-3',
        'davinci-002',
        'gpt-3.5-turbo',
        'gpt-3.5-turbo-0125',
        'gpt-3.5-turbo-0301',
        'gpt-3.5-turbo-0613',
        'gpt-3.5-turbo-1106',
        'gpt-3.5-turbo-16k',
        'gpt-3.5-turbo-16k-0613',
        'gpt-3.5-turbo-instruct',
        'gpt-3.5-turbo-instruct-0914',
        'gpt-4',
        'gpt-4-0125-preview',
        'gpt-4-0613',
        'gpt-4-1106-preview',
        'gpt-4-1106-vision-preview',
        'gpt-4-turbo',
        'gpt-4-turbo-2024-04-09',
        'gpt-4-turbo-preview',
        'gpt-4-vision-preview',
        'gpt-4o',
        'gpt-4o-2024-05-13',
        'text-embedding-3-large',
        'text-embedding-3-small',
        'text-embedding-ada-002',
        'tts-1',
        'tts-1-1106',
        'tts-1-hd',
        'tts-1-hd-1106',
        'whisper-1',
    ]

    api_key = config.api_key

    st.subheader("Key Level")

    level_model_map = {}
    api_key.level: dict
    # sort api_key.level by key, so that level 1 is always the first one
    sorted_list = sorted(api_key.level.items(), key=lambda x: x[0], reverse=False)
    levels = [i[0] for i in sorted_list]
    models_list = [i[1] for i in sorted_list]
    num_levels = st.number_input(
        '请选择你需定义几个level (不包含level 0,它是全权限)', 0, 1000, len(levels), 1
    )
    for level in range(num_levels):
        level += 1
        models = st.multiselect(
            f'请选择该`{level=}`可访问的模型',
            openai_model_list,
            default=models_list[min(level - 1, len(levels) - 1)],
            key=f"level_{level}",
        )
        level_model_map[level] = models

    with st.form("api_key_form", border=False):
        st.subheader("OpenAI API Key")
        df = pd.DataFrame(
            [
                {'api_key': key, 'level': value}
                for key, value in api_key.openai_key.items()
            ]
        )
        edited_df = st.data_editor(
            df, num_rows="dynamic", key="editor1", use_container_width=True
        )

        st.subheader("Forward Key")
        df2 = pd.DataFrame(
            [
                {'api_key': key, 'level': value}
                for key, value in api_key.forward_key.items()
            ]
        )
        edited_df2 = st.data_editor(
            df2, num_rows="dynamic", key="editor2", use_container_width=True
        )

        submitted = st.form_submit_button("Save", use_container_width=True)
        if submitted:
            api_key.openai_key = {
                row["api_key"]: row["level"] for i, row in edited_df.iterrows()
            }

            api_key.forward_key = {
                row["api_key"]: row["level"] for i, row in edited_df2.iterrows()
            }

            api_key.level = level_model_map

            print(api_key.convert_to_env())


def display_cache_configuration():
    cache = config.cache

    with st.container():
        st.subheader("Cache Configuration")

        cache_openai = st.checkbox("Cache OpenAI route", cache.cache_openai)
        cache_default_request_caching_value = st.checkbox(
            "For OpenAI API, return using cache by default",
            cache.default_request_caching_value,
            disabled=not cache_openai,
        )

        cache_general = st.checkbox("Cache General route", cache.cache_general)

        cache_backend = st.selectbox(
            "Cache Backend",
            ["MEMORY", "LMDB", "LevelDB"],
            index=["memory", "lmdb", "leveldb"].index(cache.backend.lower()),
        )

        cache_root_path_or_url = st.text_input(
            "Root Path or URL",
            cache.root_path_or_url,
            disabled=cache_backend == "MEMORY",
        )

        df = pd.DataFrame([{"cache_route": i} for i in cache.cache_routes])
        edited_df = st.data_editor(
            df, num_rows="dynamic", key="editor1", use_container_width=True
        )

        submitted = st.button("Save", use_container_width=True)
        if submitted:
            cache.cache_openai = cache_openai
            cache.cache_general = cache_general

            cache.backend = cache_backend
            cache.root_path_or_url = cache_root_path_or_url
            cache.default_request_caching_value = cache_default_request_caching_value

            cache.cache_routes = [
                row['cache_route']
                for i, row in edited_df.iterrows()
                if row["cache_route"] is not None
            ]

            print(cache.convert_to_env())


def display_rate_limit_configuration():
    rate_limit = config.rate_limit

    with st.form("rate_limit_form", border=False):
        st.subheader("Rate Limit Configuration")

        global_rate_limit = st.text_input(
            "Global Rate Limit", rate_limit.global_rate_limit
        )
        iter_chunk = st.selectbox(
            "Iteration Chunk Type",
            ['one-by-one', 'efficiency'],
            index=['one-by-one', 'efficiency'].index(rate_limit.iter_chunk),
        )
        strategy = st.selectbox(
            "Rate Limit Strategy",
            ['fixed-window', 'moving-window', 'fixed-window-elastic-expiry'],
            index=[
                'fixed-window',
                'moving-window',
                'fixed-window-elastic-expiry',
            ].index(rate_limit.strategy),
        )

        st.subheader("Token Rate Limit")
        token_rate_limit_df = pd.DataFrame(
            [i.to_dict_str() for i in rate_limit.token_rate_limit]
        )
        edited_token_rate_limit_df = st.data_editor(
            token_rate_limit_df,
            num_rows="dynamic",
            key="editor_token_rate_limit",
            use_container_width=True,
        )

        st.subheader("Request Rate Limit")
        req_rate_limit_df = pd.DataFrame(
            [i.to_dict_str() for i in rate_limit.req_rate_limit]
        )
        edited_req_rate_limit_df = st.data_editor(
            req_rate_limit_df,
            num_rows="dynamic",
            key="editor_req_rate_limit",
            use_container_width=True,
        )

        if st.form_submit_button("Save", use_container_width=True):
            rate_limit.global_rate_limit = global_rate_limit
            rate_limit.iter_chunk = iter_chunk
            rate_limit.strategy = strategy

            rate_limit.token_rate_limit = [
                RateLimitType(row["route"], ast.literal_eval(row["value"]))
                for _, row in edited_token_rate_limit_df.iterrows()
            ]

            rate_limit.req_rate_limit = [
                RateLimitType(row["route"], ast.literal_eval(row["value"]))
                for _, row in edited_req_rate_limit_df.iterrows()
            ]

            print(rate_limit.convert_to_env())


def display_other_configuration():
    with st.form("other_form", border=False):
        st.subheader("Other Configuration")

        timezone = st.text_input("Timezone", config.timezone)
        timeout = st.number_input(
            "Timeout (seconds)", value=int(config.timeout), min_value=1, step=1
        )
        proxy = st.text_input(
            "Proxy", config.proxy, placeholder="e.g. http://127.0.0.1:7890"
        )
        benchmark_mode = st.checkbox("Benchmark Mode", config.benchmark_mode)

        if st.form_submit_button("Save", use_container_width=True):
            config.timezone = timezone
            config.timeout = timeout
            config.proxy = proxy
            config.benchmark_mode = benchmark_mode

            print(config.convert_to_env())


if selected_section == "Forward":
    display_forward_configuration()

elif selected_section == "API Key && Level":
    display_api_key_configuration()

elif selected_section == "Cache":
    display_cache_configuration()

elif selected_section == "Rate Limit":
    display_rate_limit_configuration()

elif selected_section == "Other":
    display_other_configuration()

elif selected_section == "Real-time Logs":
    st.write("### Real-time Logs")
    st.write("\n")
    render_with_markdown = st.toggle("Using Markdown to render", value=True)
    st.write("\n")

    q = st.session_state['q']
    chat_data: ChatData = st.session_state['chat_data']
    chat_data.render_all_messages(markdown=render_with_markdown)
    while True:
        uid, msg = q.get()
        uid: bytes
        item = orjson.loads(msg)
        item['uid'] = uid[1:].decode()
        if uid.startswith(b"0"):
            item['user_role'] = True
        else:
            item['assistant_role'] = True

        chat_data.add_message(item, markdown=render_with_markdown)

elif selected_section == "Playground":
    st.write("## todo")

elif selected_section == "Statistics":
    st.write("## todo")



import atexit
import os
import pickle
import platform
import signal
import subprocess

import fire
import uvicorn


class Cli:
    def run_web(self, port=8001, openai_forward_host='localhost', wait=True):
        """
        Runs the web UI using the Streamlit server.

        Args:
            port (int): The port number on which to run the server.
            openai_forward_host (str): The host of the OpenAI Forward server.
            wait (bool): Whether to wait for the server to stop. Default is True.

        Returns:
            None
        """
        os.environ['OPENAI_FORWARD_HOST'] = openai_forward_host
        try:
            self._start_streamlit(port=port, wait=wait)
        except KeyboardInterrupt:
            ...
        except Exception as e:
            raise

    def run(self, port=8000, workers=1, webui=False, start_ui=True, ui_port=8001):
        """
        Runs the application using the Uvicorn server.

        Args:
            port (int): The port number on which to run the server.
            workers (int): The number of worker processes to run.
            webui (bool): Whether to run the web UI. Default is False.
            start_ui (bool): Whether to start the web UI.
            ui_port (int): The port number on which to run streamlit.

        Returns:
            None
        """

        if platform.system() == "Windows":
            os.environ["TZ"] = ""

        ssl_keyfile = os.environ.get("ssl_keyfile", None) or None
        ssl_certfile = os.environ.get("ssl_certfile", None) or None

        if not webui:
            uvicorn.run(
                app="openai_forward.app:app",
                host="0.0.0.0",
                port=port,
                workers=workers,
                app_dir="..",
                ssl_keyfile=ssl_keyfile,
                ssl_certfile=ssl_certfile,
            )
        else:
            import threading

            import zmq

            os.environ['OPENAI_FORWARD_WEBUI'] = 'true'

            context = zmq.Context()
            socket = context.socket(zmq.REP)
            restart_port = int(os.environ.get('WEBUI_RESTART_PORT', 15555))
            socket.bind(f"tcp://*:{restart_port}")
            log_socket = context.socket(zmq.ROUTER)
            log_port = int(os.environ.get("WEBUI_LOG_PORT", 15556))
            log_socket.bind(f"tcp://*:{log_port}")
            subscriber_info = {}

            def mq_worker(log_socket: zmq.Socket):

                while True:
                    identity, uid, message = log_socket.recv_multipart()
                    if uid == b"/subscribe":
                        subscriber_info[identity] = True
                        continue
                    else:
                        for subscriber, _ in subscriber_info.items():
                            log_socket.send_multipart([subscriber, uid, message])

            thread = threading.Thread(target=mq_worker, args=(log_socket,))
            thread.daemon = True
            thread.start()

            self._start_uvicorn(
                port=port,
                workers=workers,
                ssl_keyfile=ssl_keyfile,
                ssl_certfile=ssl_certfile,
            )

            if start_ui:
                self._start_streamlit(port=ui_port, wait=False)

            atexit.register(self._stop_uvicorn)

            while True:
                message = socket.recv()
                env_dict: dict = pickle.loads(message)

                for key, value in env_dict.items():
                    os.environ[key] = value

                self._restart_uvicorn(
                    port=port,
                    workers=workers,
                    ssl_keyfile=ssl_keyfile,
                    ssl_certfile=ssl_certfile,
                )
                socket.send(f"Restart success!".encode())

    def _start_uvicorn(self, port, workers, ssl_keyfile=None, ssl_certfile=None):
        from openai_forward.helper import wait_for_serve_start

        self.uvicorn_proc = subprocess.Popen(
            [
                'uvicorn',
                'openai_forward.app:app',
                '--host',
                '0.0.0.0',
                '--port',
                str(port),
                '--app-dir',
                '..',
                '--workers',
                str(workers),
            ]
            + (['--ssl-keyfile', ssl_keyfile] if ssl_keyfile else [])
            + (['--ssl-certfile', ssl_certfile] if ssl_certfile else [])
        )
        suppress_exception = platform.system() == "Windows"
        wait_for_serve_start(
            f"http://localhost:{port}/healthz",
            timeout=10,
            suppress_exception=suppress_exception,
        )

    def _start_streamlit(self, port, wait=False):
        from openai_forward.helper import relp

        self.streamlit_proc = subprocess.Popen(
            [
                'streamlit',
                'run',
                f'{relp("webui/run.py")}',
                '--server.port',
                str(port),
                '--server.headless',
                'true',
                '--server.enableCORS',
                'true',
                '--server.runOnSave',
                'true',
                '--theme.base',
                'light',
                '--browser.gatherUsageStats',
                'false',
            ]
        )

        atexit.register(self._stop_streamlit)
        if wait:
            self.streamlit_proc.wait()

    def _restart_uvicorn(self, **kwargs):
        self._stop_uvicorn()
        self._start_uvicorn(**kwargs)

    def _stop_streamlit(self):
        self._stop(uvicorn=False)

    def _stop_uvicorn(self):
        self._stop(streamlit=False)

    def _stop(self, uvicorn=True, streamlit=True):
        if uvicorn and self.uvicorn_proc.poll() is None:
            self.uvicorn_proc.send_signal(signal.SIGINT)
            try:
                self.uvicorn_proc.wait(timeout=15)
            except subprocess.TimeoutExpired:
                self.uvicorn_proc.kill()
        if streamlit and self.streamlit_proc.poll() is None:
            self.streamlit_proc.send_signal(signal.SIGINT)
            try:
                self.streamlit_proc.wait(timeout=5)
            except subprocess.TimeoutExpired:
                self.streamlit_proc.kill()

    @staticmethod
    def convert(log_folder: str = None, target_path: str = None):
        """
        Converts log files in a folder to a JSONL file.

        Args:
            log_folder (str, optional): The path to the folder containing the log files. Default is None.
            target_path (str, optional): The path to the target JSONL file. Default is None.

        Returns:
            None
        """
        from openai_forward.helper import convert_folder_to_jsonl, route_prefix_to_str
        from openai_forward.settings import OPENAI_ROUTE_PREFIX

        print(60 * '-')
        if log_folder is None:
            if target_path is not None:
                raise ValueError("target_path must be None when log_folder is None")
            _prefix_list = [route_prefix_to_str(i) for i in OPENAI_ROUTE_PREFIX]
            for prefix in _prefix_list:
                log_folder = f"./Log/{prefix}/chat"
                target_path = f"./Log/chat_{prefix}.json"
                print(f"Convert {log_folder}/*.log to {target_path}")
                convert_folder_to_jsonl(log_folder, target_path)
                print(60 * '-')
        else:
            print(f"Convert {log_folder}/*.log to {target_path}")
            convert_folder_to_jsonl(log_folder, target_path)
            print(60 * '-')

    @staticmethod
    def gen_config(dir: str = "."):
        """
        Generates a .env file in the specified directory.
        """
        from pathlib import Path

        from openai_forward.config.interface import Config

        config = Config()
        env_dict = config.convert_to_env(set_env=False)
        dir = Path(dir)

        with open(dir / ".env", "w") as f:
            env_content = "\n".join(
                [f"{key}={value}" for key, value in env_dict.items()]
            )
            f.write(env_content)


def main():
    fire.Fire(Cli)


if __name__ == "__main__":
    main()



import streamlit as st
from streamlit.components.v1 import components


def mermaid(code: str) -> None:
    st.components.v1.html(
        f"""
            <pre class="mermaid">
                {code}
            </pre>

            <script type="module">
                import mermaid from 'https://cdn.jsdelivr.net/npm/mermaid@10/dist/mermaid.esm.min.mjs';
                mermaid.initialize({{ startOnLoad: true }});
            </script>
            """,
        height=300,
    )



from flaxkv import FlaxKV

from ..settings import CACHE_BACKEND, CACHE_ROOT_PATH_OR_URL, LOG_CACHE_DB_INFO

if CACHE_BACKEND.upper() == "MEMORY":
    db_dict = {}

elif CACHE_BACKEND.lower() in ("leveldb", "lmdb"):

    log, save_log = ("INFO", True) if LOG_CACHE_DB_INFO else (None, False)

    db_dict = FlaxKV(
        "CACHE_DB",
        root_path_or_url=CACHE_ROOT_PATH_OR_URL,
        backend=CACHE_BACKEND.lower(),
        cache=True,
        log=log,
        save_log=save_log,
    )

else:
    raise ValueError(
        f"Unknown cache backend: {CACHE_BACKEND}. The valid backends are: 'leveldb', 'lmdb', 'memory'"
    )



from __future__ import annotations

from fastapi.responses import Response
from flaxkv.pack import encode
from loguru import logger

from ...settings import CACHE_OPENAI
from ..database import db_dict


def construct_cache_key(payload_info):
    elements = [
        payload_info['model'],
        payload_info['input'],
        payload_info['encoding_format'],
    ]
    return encode(elements)


def get_cached_embedding_response(payload_info, valid_payload, request, **kwargs):
    """
    Attempts to retrieve a cached response based on the current request's payload information.

    Note:
        If a cache hit occurs, the cached response is immediately returned without contacting the external server.
    """
    if not (CACHE_OPENAI and valid_payload):
        return None, None

    cache_key = construct_cache_key(payload_info)

    if payload_info['caching'] and cache_key in db_dict:
        logger.info(f'embedding uid: {payload_info["uid"]} >>>>> [cache hit]')
        cache_data = db_dict[cache_key]
        content = cache_data['data']

        return Response(content=content, media_type="application/json"), cache_key

    return None, cache_key



from typing import List

from rich import print
from rich.panel import Panel
from rich.table import Table

from . import __version__


def print_startup_info(base_url, route_prefix, api_key, fwd_key, /, style, **kwargs):
    """
    Prints the startup information of the application.
    """
    try:
        from dotenv import load_dotenv

        load_dotenv(".env")
    except Exception:
        ...
    route_prefix = route_prefix or "/"
    if not isinstance(api_key, str):
        api_key = True if len(api_key) else False
    if not isinstance(fwd_key, str):
        fwd_key = True if len(fwd_key) else False
    table = Table(title="", box=None, width=61)

    metric = {
        "base url": {
            'value': base_url,
        },
        "route prefix": {
            'value': route_prefix,
        },
        "api keys": {
            'value': str(api_key),
        },
        "forward keys": {
            'value': str(fwd_key),
            'style': "#62E883" if fwd_key or not api_key else "red",
        },
    }
    table.add_column("", justify='left', width=12)
    table.add_column("", justify='left')
    for key, value in metric.items():
        if value['value']:
            table.add_row(key, value['value'], style=value.get('style', style))
    for key, value in kwargs.items():
        if value:
            table.add_row(key, str(value), style=style)

    print(
        Panel(
            table,
            title=f"🤗 openai-forward (v{__version__}) is ready to serve! ",
            expand=False,
        )
    )


def print_rate_limit_info(
    backend: str,
    strategy: str,
    global_req_rate_limit: str,
    req_rate_limit: dict,
    token_rate_limit: dict,
    **kwargs,
):
    """
    Print rate limit information.
    """
    table = Table(title="", box=None, width=61)
    table.add_column("")
    table.add_column("", justify='left')
    backend = backend or "memory"
    table.add_row("backend", backend, style='#7CD9FF')
    if strategy:
        table.add_row("strategy", strategy, style='#7CD9FF')

    if global_req_rate_limit:
        table.add_row(
            "global rate limit", f"{global_req_rate_limit} (req)", style='#C5FF95'
        )
    for key, value in req_rate_limit.items():
        table.add_row(key, f"{value} (req)", style='#C5FF95')

    for key, value in token_rate_limit.items():
        if isinstance(value, float):
            value = f"{value:.4f} s/token"
        table.add_row(key, f"{value} (token)", style='#C5FF95')

    for key, value in kwargs.items():
        table.add_row(key, str(value), style='#C5FF95')

    print(Panel(table, title="⏱️ Rate Limit configuration", expand=False))



import os
import threading
import time
from abc import ABC, abstractmethod
from typing import Dict, List, Tuple

import orjson
from fastapi import Request
from loguru import logger
from orjson import JSONDecodeError

from ..helper import get_client_ip, get_unique_id, route_prefix_to_str
from ..settings import DEFAULT_REQUEST_CACHING_VALUE
from .helper import markdown_print, parse_sse_buffer, print


class LoggerBase(ABC):
    def __init__(self, route_prefix: str, _suffix: str):
        _prefix = route_prefix_to_str(route_prefix)
        kwargs = {f"{_prefix}{_suffix}": True}
        self.logger = logger.bind(**kwargs)

        self.webui = False
        if os.environ.get("OPENAI_FORWARD_WEBUI", "false").strip().lower() == 'true':
            self.webui = True

            import zmq
            from flaxkv.helper import SimpleQueue

            context = zmq.Context()
            socket = context.socket(zmq.DEALER)
            webui_log_port = int(os.environ.get("WEBUI_LOG_PORT", 15556))
            socket.connect(f"tcp://localhost:{webui_log_port}")

            self.q = SimpleQueue(maxsize=200)

            def _worker():
                while True:
                    message: dict = self.q.get(block=True)
                    if message.get("payload"):
                        uid = b"0" + message["uid"].encode()
                        msg = message["payload"]
                    else:
                        uid = b"1" + message["uid"].encode()
                        msg = message["result"]
                        msg = orjson.dumps(msg)

                    socket.send_multipart([uid, msg])

            threading.Thread(target=_worker, daemon=True).start()

    @staticmethod
    @abstractmethod
    def parse_payload(request: Request, raw_payload) -> Tuple[Dict, bytes]:
        pass

    @staticmethod
    @abstractmethod
    def parse_bytearray(buffer: bytearray) -> Dict:
        pass

    @abstractmethod
    def log_result(self, *args, **kwargs):
        pass


class CompletionLogger(LoggerBase):
    def __init__(self, route_prefix: str):
        """
        Initialize the Completions logger with a route prefix.

        Args:
            route_prefix (str): The prefix used for routing, e.g., '/openai'.
        """
        super().__init__(route_prefix, "_completion")

    @staticmethod
    def parse_payload(request: Request, raw_payload):
        uid = get_unique_id()
        payload = orjson.loads(raw_payload)

        content = {
            "prompt": payload['prompt'],
            "model": payload['model'],
            "stream": payload.get("stream", True),
            "temperature": payload.get("temperature", 1),
            "ip": get_client_ip(request) or "",
            "uid": uid,
            "datetime": time.strftime("%Y-%m-%d %H:%M:%S", time.localtime()),
        }
        return content, raw_payload

    @staticmethod
    def parse_bytearray(buffer: bytearray):
        """
        Parses a bytearray, usually from an API response, into a dictionary containing various information.

        Args:
            buffer (List[bytes]): A list of bytes to parse.

        Returns:
            Dict[str, Any]: A dictionary containing metadata and content. The keys include:
                - "text" (str)
        """
        start_token = "data: "
        start_token_len = len(start_token)
        if buffer.startswith(b'data: '):
            txt_lines = parse_sse_buffer(buffer)
            stream = True
            first_dict = orjson.loads(txt_lines[0][start_token_len:])
            text = first_dict["choices"][0]["text"]
        else:
            stream = False
            first_dict = orjson.loads(buffer)
            text = first_dict["choices"][0]["text"]

        target_info = dict()
        verbose = False
        if verbose:
            target_info["created"] = first_dict["created"]
            target_info["id"] = first_dict["id"]
            target_info["model"] = first_dict["model"]
            target_info["usage"] = first_dict.get("usage")

        target_info['text'] = text

        if not stream:
            return target_info

        # loop for stream
        for line in txt_lines[1:]:
            if line.startswith(start_token):
                try:
                    line_dict = orjson.loads(line[start_token_len:])
                    text += line_dict["choices"][0]["text"]
                except (JSONDecodeError, KeyError):
                    pass
        target_info['text'] = text
        return target_info

    def log_result(self, chat_info: dict):
        """
        Log chat information to the logger bound to this instance.

        Args:
            chat_info (dict): A dictionary containing chat information to be logged.
        """
        self.logger.debug(f"{chat_info}")


class ChatLogger(LoggerBase):
    def __init__(self, route_prefix: str):
        """
        Initialize the Chat Completions logger with a route prefix.

        Args:
            route_prefix (str): The prefix used for routing, e.g., '/openai'.
        """
        super().__init__(route_prefix, "_chat")

    def parse_payload(self, request: Request, raw_payload):
        """
        Asynchronously parse the payload from a FastAPI request.

        Args:
            request (Request): A FastAPI request object.

        Returns:
            dict: A dictionary containing parsed messages, model, IP address, UID, and datetime.
        """
        uid = get_unique_id()

        if self.webui:
            self.q.put({"uid": uid, "payload": raw_payload})

        payload = orjson.loads(raw_payload)
        caching = payload.pop("caching", None)
        if caching is None:
            caching = DEFAULT_REQUEST_CACHING_VALUE
            payload_return = raw_payload
        else:
            payload_return = orjson.dumps(payload)

        info = {
            "messages": payload["messages"],
            "model": payload["model"],
            "stream": payload.get("stream", False),
            "max_tokens": payload.get("max_tokens", None),
            "response_format": payload.get("response_format", None),
            "n": payload.get("n", 1),
            "temperature": payload.get("temperature", 1),
            "top_p": payload.get("top_p", 1),
            "logit_bias": payload.get("logit_bias", None),
            "frequency_penalty": payload.get("frequency_penalty", 0),
            "presence_penalty": payload.get("presence_penalty", 0),
            "seed": payload.get("seed", None),
            "stop": payload.get("stop", None),
            "user": payload.get("user", None),
            "tools": payload.get("tools", None),
            "tool_choice": payload.get("tool_choice", None),
            "ip": get_client_ip(request) or "",
            "uid": uid,
            "caching": caching,
            "datetime": time.strftime("%Y-%m-%d %H:%M:%S", time.localtime()),
        }

        return info, payload_return

    def parse_bytearray(self, buffer: bytearray):
        """
        Parses a bytearray, usually from an API response, into a dictionary containing various information.

        Args:
            buffer (List[bytes]): A list of bytes to parse.

        Returns:
            Dict[str, Any]: A dictionary containing metadata and content. The keys include:
                - "assistant" (str): content
                - "is_tool_calls" (boolean)
        """
        start_token = "data: "
        start_token_len = len(start_token)
        if buffer.startswith(b'data: '):
            txt_lines = parse_sse_buffer(buffer)
            stream = True
            first_dict = orjson.loads(txt_lines[0][start_token_len:])
            # todo: multiple choices
            msg = first_dict["choices"][0]["delta"]
        else:
            stream = False
            first_dict = orjson.loads(buffer)
            # todo: multiple choices
            msg = first_dict["choices"][0]["message"]

        target_info = dict()
        verbose = False
        if verbose:
            target_info["created"] = first_dict["created"]
            target_info["id"] = first_dict["id"]
            target_info["model"] = first_dict["model"]
            target_info["role"] = msg["role"]
        role = msg["role"]  # always be "assistant"

        content, tool_calls = msg.get("content"), msg.get("tool_calls")
        if tool_calls:
            """
            tool_calls:
                [{
                "index": 0,
                "id": 'xx',
                'type": 'function',
                'function': {'name': 'xxx', 'arguments': ''}
                 }]
            """
            target_info[role] = tool_calls
            target_info["is_tool_calls"] = True
            parse_content_key = "tool_calls"

        else:
            target_info[role] = content
            target_info["is_tool_calls"] = False
            parse_content_key = "content"

        if not stream:
            return target_info

        # loop for stream
        stream_content = ""
        for line in txt_lines[1:]:
            if line.startswith(start_token):
                stream_content += self._parse_one_line_content(
                    line[start_token_len:], parse_content_key
                )
        if target_info['is_tool_calls']:
            tool_calls[0]['function']['arguments'] = stream_content
            target_info[role] = tool_calls
        else:
            target_info[role] = stream_content
        return target_info

    @staticmethod
    def _parse_one_line_content(line: str, parse_key="content"):
        """
        Helper method to parse a single line.

        Args:
            line (str): The line to parse.
            parse_key (str): .

        Returns:
            str: The parsed content from the line.
        """
        try:
            line_dict = orjson.loads(line)
            if parse_key == "content":
                return line_dict["choices"][0]["delta"][parse_key]
            elif parse_key == "tool_calls":
                tool_calls = line_dict["choices"][0]["delta"]["tool_calls"]
                return tool_calls[0]["function"]['arguments']
            else:
                logger.error(f"Unknown parse key: {parse_key}")
                return ""
        except JSONDecodeError:
            return ""
        except KeyError:
            return ""

    def log(self, chat_info: dict):
        """
        Log chat information to the logger bound to this instance.

        Args:
            chat_info (dict): A dictionary containing chat information to be logged.
        """
        self.logger.debug(f"{chat_info}")

    def log_result(self, chat_info: dict):
        """
        Log chat information to the logger bound to this instance.

        Args:
            chat_info (dict): A dictionary containing chat information to be logged.
        """
        self.logger.debug(f"{chat_info}")

    @staticmethod
    def print_chat_info(chat_info: dict):
        """
        Print chat information in a formatted manner.

        Args:
            chat_info (dict): A dictionary containing chat information to be printed.
        """
        messages = chat_info.get("messages")
        if messages:
            print(50 * "-", role='other')
            for msg in messages:
                for key, value in msg.items():
                    markdown_print(f"{key}: {value}", role=key)
            print(
                f"{chat_info.get('ip')}@{chat_info.get('model')} uid: {chat_info.get('uid')}",
                role='other',
            )
            print(50 * "-", role='other')
        else:
            assistant = chat_info.get("assistant")
            if assistant:
                print(77 * "=", role='assistant')
                markdown_print(f"assistant: {assistant}", role='assistant')
                print(f'uid: {chat_info.get("uid")}')
                print(77 * "=", role='assistant')


class EmbeddingLogger(LoggerBase):
    def __init__(self, route_prefix: str):
        super().__init__(route_prefix, "_embedding")

    @staticmethod
    def parse_payload(request: Request, raw_payload: bytes):
        uid = get_unique_id()
        payload = orjson.loads(raw_payload)
        caching = payload.pop("caching", None)
        if caching is None:
            caching = DEFAULT_REQUEST_CACHING_VALUE
            payload_return = raw_payload
        else:
            payload_return = orjson.dumps(payload)

        content = {
            "input": payload['input'],
            "model": payload['model'],
            "encoding_format": payload.get("encoding_format", 'float'),
            "ip": get_client_ip(request) or "",
            "uid": uid,
            "caching": caching,
            "datetime": time.strftime("%Y-%m-%d %H:%M:%S", time.localtime()),
        }
        return content, payload_return

    def parse_bytearray(self, buffer: bytearray):
        """
        Parse a bytearray into a dictionary.
        """
        result_dict = orjson.loads(buffer)
        target_info = {
            "object": result_dict["object"],
            "usage": result_dict["usage"],
            "model": result_dict["model"],
            "buffer": buffer,
        }
        return target_info

    def log(self, info: dict):
        self.logger.debug(f"{info}")

    def log_result(self, info: dict):
        result_info = {
            "object": info["object"],
            "usage": info["usage"],
            "model": info["model"],
            "uid": info["uid"],
        }
        self.logger.debug(f"{result_info}")


class WhisperLogger:
    # todo
    def __init__(self, route_prefix: str):
        """
        Initialize the Audio logger with a route prefix.

        Args:
            route_prefix (str): The prefix used for routing, e.g., '/openai'.
        """
        _prefix = route_prefix_to_str(route_prefix)
        self.logger = logger.bind(**{f"{_prefix}_whisper": True})

    def log_buffer(self, bytes_: bytes):
        """
        Add a log entry containing the decoded text from bytes.

        Args:
            bytes_ (bytes): The byte string to decode and log.
        """
        text_content = bytes_.decode("utf-8")
        self.logger.info(text_content)



from __future__ import annotations

import asyncio
import traceback
from asyncio import Queue
from itertools import cycle
from typing import Any, AsyncGenerator, Iterable

import aiohttp
import anyio
from aiohttp import TCPConnector
from fastapi import HTTPException, Request, status
from loguru import logger
from starlette.responses import BackgroundTask, StreamingResponse

from ..cache import (
    cache_generic_response,
    cache_response,
    get_cached_generic_response,
    get_cached_response,
)
from ..content.openai import (
    ChatLogger,
    CompletionLogger,
    EmbeddingLogger,
    WhisperLogger,
)
from ..decorators import async_retry, async_token_rate_limit_auth_level
from ..helper import InfiniteSet, get_client_ip, get_unique_id
from ..settings import *

# from beartype import beartype


class GenericForward:
    """
    Base class for handling request forwarding to another service.
    Provides methods for request validation, logging, and proxying.
    """

    validate_host = bool(IP_BLACKLIST or IP_WHITELIST)
    timeout = aiohttp.ClientTimeout(connect=TIMEOUT)

    _fk_to_level = FWD_KEY
    _sk_to_levels = OPENAI_API_KEY
    _level_to_model_set = {level: set(models) for level, models in LEVEL_MODELS.items()}

    _zero_level_model_set = InfiniteSet()

    _level_to_sks = {}
    for sk, levels in _sk_to_levels.items():
        for level in levels:
            _level_to_sks[level] = _level_to_sks.get(level, []) + [sk]

    _level_to_sk = {}
    for level, sks in _level_to_sks.items():
        _level_to_sk[level] = cycle(sks)

    def __init__(self, base_url: str, route_prefix: str, proxy=None):
        """
        Args:
            base_url (str): The base URL to which requests will be forwarded.
            route_prefix (str): The prefix of the route.
            proxy (str, optional): The proxy to use for the requests. Defaults to None.
        """
        self.BASE_URL = base_url
        self.PROXY = proxy
        self.ROUTE_PREFIX = route_prefix
        self.client: aiohttp.ClientSession | None = None

    @classmethod
    def fk_to_sk(cls, forward_key: str):
        """
        Convert a forward key to a secret key.

        Args:
            forward_key (str): The forward key to convert.

        Returns:
            str: The corresponding secret key, if it exists. Otherwise, None.
        """
        level = cls._fk_to_level.get(forward_key)
        if level is not None:
            sk = cls._level_to_sk.get(level)
            if sk:
                return next(sk), level
        return None, level

    @classmethod
    def handle_authorization(cls, client_config):
        """
        Handle the authorization for the client.

        Args:
            client_config (dict): The configuration for the client.

        Returns:
            str: The authorization string.
            set: The set of models can be accessed.
        """
        auth, auth_prefix = client_config["auth"], "Bearer "
        model_set = cls._zero_level_model_set

        if auth:
            fk = auth[len(auth_prefix) :]
            if fk in FWD_KEY:
                sk, level = cls.fk_to_sk(fk)
                assert level is not None
                if level != 0:
                    model_set = cls._level_to_model_set[level]
                if sk:
                    auth = auth_prefix + sk
                    client_config["headers"]["Authorization"] = auth
        return auth, model_set

    async def build_client(self):
        """
        Asynchronously build the client for making requests.
        """
        connector = TCPConnector(limit=500, limit_per_host=0, force_close=False)
        self.client = aiohttp.ClientSession(connector=connector, timeout=self.timeout)

    @staticmethod
    def validate_request_host(ip):
        """
        Validates the incoming IP address against a whitelist and blacklist.

        Args:
            ip (str): The IP address to validate.

        Raises:
            HTTPException: If the IP is not in the whitelist or if it is in the blacklist.
        """
        if (
            IP_WHITELIST
            and ip not in IP_WHITELIST
            or IP_BLACKLIST
            and ip in IP_BLACKLIST
        ):
            logger.warning(f"IP {ip} is unauthorized")
            raise HTTPException(
                status_code=status.HTTP_403_FORBIDDEN,
                detail=f"Forbidden Error",
            )

    @staticmethod
    @async_token_rate_limit_auth_level(token_interval_conf, FWD_KEY)
    async def aiter_bytes(
        r: aiohttp.ClientResponse,
        request: Request,
        route_path: str,
        cache_key: str | None = None,
    ) -> AsyncGenerator[bytes, Any]:
        """
        Asynchronously iterates through the bytes in the given aiohttp.ClientResponse object
        and yields each chunk while also caching the response if necessary.

        Args:
            r (aiohttp.ClientResponse): The aiohttp.ClientResponse object.
            request (Request): The original FastAPI request object.
            route_path (str): The API route path.
            cache_key (str | None): The cache key. Defaults to None.

        Returns:
            AsyncGenerator[bytes, Any]: Each chunk of bytes from the server's response.
        """
        chunk_list = []
        cache = True if route_path in CACHE_ROUTE_SET else False

        async for chunk, _ in r.content.iter_chunks():  # yield chunk one by one.
            if cache:
                chunk_list.append(chunk)
            yield chunk

        if r.ok and cache and cache_key:
            cache_generic_response(cache_key, chunk_list, route_path)

        # Only log non-stream response:
        if len(chunk_list) == 1 and LOG_GENERAL:
            logger.debug(f"result: {chunk_list[0]}")

    @async_retry(
        max_retries=3,
        delay=0.2,
        backoff=2,
        exceptions=(
            aiohttp.ServerTimeoutError,
            aiohttp.ServerConnectionError,
            aiohttp.ServerDisconnectedError,
            asyncio.TimeoutError,
            anyio.EndOfStream,
            RuntimeError,
        ),
        # raise_callback_name="build_client",
        raise_handler_name="handle_exception",
    )
    async def send(self, client_config: dict, data=None):
        """
        Asynchronously send the request and return the response.

        Args:
            client_config (dict): The configuration for the client.
            data (Any, optional): The data to send in the request. Defaults to None.

        Returns:
            aiohttp.ClientResponse: The response from the server.
        """

        return await self.client.request(
            method=client_config["method"],
            url=client_config['url'],
            data=data,
            headers=client_config["headers"],
            proxy=self.PROXY,
        )

    def handle_exception(self, e):
        """
        Handle exceptions that occur during the request.

        Args:
            e (Exception): The exception that occurred.

        Raises:
            HTTPException: An HTTPException with the appropriate status code and detail.
        """
        if isinstance(
            e,
            (
                asyncio.TimeoutError,
                aiohttp.ServerConnectionError,
                aiohttp.ServerDisconnectedError,
                aiohttp.ServerTimeoutError,
            ),
        ):
            error_info = (
                f"{type(e)}: {e} | "
                f"Please check if your host can access [{self.BASE_URL}] successfully?"
            )
            status_code = status.HTTP_504_GATEWAY_TIMEOUT

        elif isinstance(e, anyio.EndOfStream):
            error_info = "EndOfStream Error: trying to read from a stream that has been closed from the other end."
            status_code = status.HTTP_500_INTERNAL_SERVER_ERROR

        else:
            error_info = f"{type(e)}: {e}"
            status_code = status.HTTP_500_INTERNAL_SERVER_ERROR

        logger.error(f"{error_info}\n{traceback.format_exc()}")
        raise HTTPException(status_code=status_code, detail=error_info)

    def prepare_client(self, request: Request, return_origin_header=False) -> dict:
        """
        Prepare the client for making a request.

        Args:
            request (Request): The original FastAPI request object.
            return_origin_header (bool, optional): Whether to return the original header. Defaults to False.

        Returns:
            dict: The configuration for the client.
        """
        assert self.BASE_URL and self.ROUTE_PREFIX
        ip = get_client_ip(request)
        if self.validate_host:
            self.validate_request_host(ip)

        _url_path = f"{request.scope.get('root_path')}{request.scope.get('path')}"
        route_path = (
            _url_path[len(self.ROUTE_PREFIX) :]
            if self.ROUTE_PREFIX != '/'
            else _url_path
        )
        if request.url.query:
            url = f"{self.BASE_URL}{route_path}?{request.url.query}"
        else:
            url = f"{self.BASE_URL}{route_path}"

        auth = request.headers.get("Authorization", "")

        if return_origin_header:
            headers = dict(request.headers)
            headers_to_remove = [
                "host",
                "cookie",
                # "user-agent",
                # "connection",
                # "cache-control", "upgrade-insecure-requests",
                # "sec-fetch-site", "sec-fetch-mode", "sec-fetch-user", "sec-fetch-dest",
                "accept-encoding",
                "accept-language",
            ]
            for key in headers_to_remove:
                headers.pop(key, None)

        else:
            headers = {
                "content-type": request.headers.get("content-type", "application/json"),
                "authorization": auth,
            }
            for key, value in request.headers.items():
                if key.startswith("openai"):
                    headers[key] = value

        return {
            'auth': auth,
            'headers': headers,
            "method": request.method,
            'url': url,
            'ip': ip,
            'route_path': route_path,
        }

    async def _handle_payload(self, request: Request, route_path: str, model_set):

        if not request.method == "POST":
            return

        if route_path in (
            CHAT_COMPLETION_ROUTE,
            COMPLETION_ROUTE,
            EMBEDDING_ROUTE,
            CUSTOM_GENERAL_ROUTE,
        ):
            payload = await request.json()
            model = payload.get("model", None)

            if model is not None and model not in model_set:
                logger.warning(f"[Auth Warning] model: {model} is not allowed")
                raise HTTPException(
                    status_code=status.HTTP_403_FORBIDDEN,
                    detail=f"model: {model} is not allowed",
                )

    async def reverse_proxy(self, request: Request):
        """
        Asynchronously handle reverse proxying the incoming request.

        Args:
            request (Request): The incoming FastAPI request object.

        Returns:
            StreamingResponse: A FastAPI StreamingResponse containing the server's response.
        """
        assert self.client
        data = await request.body()

        if LOG_GENERAL:
            logger.debug(f"payload: {data}")
        client_config = self.prepare_client(request, return_origin_header=True)
        route_path = client_config["route_path"]

        _, model_set = self.handle_authorization(client_config)
        payload = await self._handle_payload(request, route_path, model_set)

        cached_response, cache_key = get_cached_generic_response(
            data, request, route_path
        )

        if cached_response:
            return cached_response

        r = await self.send(client_config, data=data)

        return StreamingResponse(
            self.aiter_bytes(r, request, cache_key),
            status_code=r.status,
            media_type=r.headers.get("content-type"),
            background=BackgroundTask(r.release),
        )


class OpenaiForward(GenericForward):
    """
    Derived class for handling request forwarding specifically for the OpenAI (Style) API.
    Inherits from the GenericForward class and adds specific functionality for the OpenAI API.
    """

    def __init__(self, base_url: str, route_prefix: str, proxy=None):
        """
        Initialize the OpenaiForward class.

        Args:
            base_url (str): The base URL to which requests will be forwarded.
            route_prefix (str): The prefix of the route.
            proxy (str, optional): The proxy to use for the requests. Defaults to None.
        """
        super().__init__(base_url, route_prefix, proxy)
        if LOG_OPENAI or PRINT_CHAT:
            self.chat_logger = ChatLogger(self.ROUTE_PREFIX)
            self.completion_logger = CompletionLogger(self.ROUTE_PREFIX)
            self.whisper_logger = WhisperLogger(self.ROUTE_PREFIX)
            self.embedding_logger = EmbeddingLogger(self.ROUTE_PREFIX)

    def _handle_result(
        self, buffer: bytearray, uid: str, route_path: str, request_method: str
    ):
        """
        Logs the result of the API call.

        Args:
            buffer (bytearray): List of bytes, usually from the API response.
            uid (str): Unique identifier for the request.
            route_path (str): API route path.
            request_method (str): HTTP method (e.g., 'GET', 'POST').

        Raises:
            Suppress all errors.
        """
        result_info = {}

        # If not configured to log or print chat, or the method is not POST, return early
        if not (LOG_OPENAI or PRINT_CHAT) or request_method != "POST":
            return result_info

        try:
            logger_instance = self.get_logger(route_path)
            if logger_instance:
                result_info = logger_instance.parse_bytearray(buffer)
                result_info["uid"] = uid

                if LOG_OPENAI:
                    if logger_instance.webui:
                        logger_instance.q.put({"uid": uid, "result": result_info})
                    logger_instance.log_result(result_info)

                # todo: Deprecated soon
                if PRINT_CHAT and logger_instance == self.chat_logger:
                    self.chat_logger.print_chat_info(result_info)

        except Exception:
            logger.warning(f"log result error:\n{traceback.format_exc()}")

        return result_info

    def get_logger(self, route_path: str):
        """
        Get which logger to use based on the route_path
        """
        if LOG_OPENAI:
            if route_path == CHAT_COMPLETION_ROUTE:
                return self.chat_logger
            elif route_path == COMPLETION_ROUTE:
                return self.completion_logger
            elif route_path == EMBEDDING_ROUTE:
                return self.embedding_logger
        return None

    async def _handle_payload(self, request: Request, route_path: str, model_set):
        """
        Asynchronously logs the payload of the API call.

        Args:
            request (Request): The original FastAPI request object.
            route_path (str): The API route path.
            model_set (set): The set of models that can be accessed.

        Returns:
            valid, payload_info, payload

        Raises:
            Suppress all errors.

        """
        payload_log_info = {"uid": None}

        payload = await request.body()

        if not (LOG_OPENAI or PRINT_CHAT) or request.method != "POST":
            return False, payload_log_info, payload

        try:

            logger_instance = self.get_logger(route_path)
            # If a logger method is determined, parse payload and log if necessary
            if logger_instance:
                payload_log_info, payload = logger_instance.parse_payload(
                    request, payload
                )

                if payload_log_info and LOG_OPENAI:
                    logger_instance.logger.debug(payload_log_info)

                if (
                    payload_log_info
                    and PRINT_CHAT
                    and logger_instance == self.chat_logger
                ):
                    self.chat_logger.print_chat_info(payload_log_info)
            else:
                ...

        except Exception as e:
            logger.warning(
                f"log chat error:\nhost:{request.client.host} method:{request.method}: {traceback.format_exc()}"
            )

        valid = True if payload_log_info['uid'] is not None else False

        if valid:
            if payload_log_info["model"] not in model_set:
                logger.warning(
                    f"[Auth Warning] model: {payload_log_info['model']} is not allowed"
                )
                raise HTTPException(
                    status_code=status.HTTP_403_FORBIDDEN,
                    detail=f"model: {payload_log_info['model']} is not allowed",
                )

        return valid, payload_log_info, payload

    @staticmethod
    async def read_chunks(r: aiohttp.ClientResponse, queue):
        """
        Read chunks of data from the response.

        Args:
            r (aiohttp.ClientResponse): The aiohttp.ClientResponse object.
            queue (Queue): The queue to put the chunks into.

        """
        buffer = bytearray()

        # Efficiency Mode
        if ITER_CHUNK_TYPE == "efficiency":
            # yield all available data as soon as it is received.
            async for chunk in r.content.iter_any():
                buffer.extend(chunk)
                await queue.put(chunk)

        # Precision Mode
        else:
            async for chunk, _ in r.content.iter_chunks():
                buffer.extend(chunk)
                await queue.put(chunk)

        await queue.put(buffer)  # add all information when the stream ends

    @async_token_rate_limit_auth_level(token_interval_conf, FWD_KEY)
    async def aiter_bytes(
        self,
        r: aiohttp.ClientResponse,
        request: Request,
        route_path: str,
        uid: str,
        cache_key: str | None = None,
        stream: bool | None = None,
    ):
        """
        Asynchronously iterates through the bytes in the given aiohttp.ClientResponse object
        and yields each chunk while also logging the request and response data.

        Args:
            r (aiohttp.ClientResponse): The aiohttp.ClientResponse object.
            request (Request): The original FastAPI request object.
            route_path (str): The API route path.
            uid (str): Unique identifier for the request.
            cache_key (bytes): The cache key.
            stream (bool): Whether the response is a stream.

        Returns:
             AsyncGenerator[bytes]: Each chunk of bytes from the server's response.
        """

        yield_completed = False
        chunk_list = []
        chunk = None
        if stream:
            queue = Queue()
            # todo:
            task = asyncio.create_task(self.read_chunks(r, queue))
            try:
                while True:
                    chunk = await queue.get()
                    if not isinstance(chunk, bytes):
                        queue.task_done()
                        yield_completed = True
                        break
                    if CACHE_OPENAI:
                        chunk_list.append(chunk)
                    yield chunk
            except Exception as e:
                logger.warning(
                    f"aiter_bytes error:{e}\nhost:{request.client.host} method:{request.method}: "
                    f"{traceback.format_exc()}"
                )
            finally:
                if not task.done():
                    task.cancel()
        else:
            try:
                chunk = await r.read()
                yield chunk
                chunk_list.append(chunk)
                chunk = bytearray(chunk)
                yield_completed = True
            except Exception as e:
                logger.warning(
                    f"aiter_bytes error:{e}\nhost:{request.client.host} method:{request.method}: "
                    f"{traceback.format_exc()}"
                )

        r.release()

        if uid:
            if r.ok and yield_completed:
                target_info = self._handle_result(
                    chunk, uid, route_path, request.method
                )
                if CACHE_OPENAI:
                    cache_response(cache_key, target_info, route_path, chunk_list)

            elif chunk is not None:
                logger.warning(
                    f'uid: {uid}\n status: {r.status}\n {chunk.decode("utf-8")}'
                )
            else:
                logger.warning(f'uid: {uid}\n' f'{r.status}')

    async def reverse_proxy(self, request: Request):
        """
        Asynchronously handles reverse proxying the incoming request.

        Args:
            request (Request): The incoming FastAPI request object.

        Returns:
            StreamingResponse: A FastAPI StreamingResponse containing the server's response.
        """
        client_config = self.prepare_client(request, return_origin_header=False)
        route_path = client_config["route_path"]

        _, model_set = self.handle_authorization(client_config)
        valid_payload, payload_info, payload = await self._handle_payload(
            request, route_path, model_set
        )
        uid = payload_info["uid"]
        stream = payload_info.get('stream', DEFAULT_STREAM_RESPONSE)

        cached_response, cache_key = get_cached_response(
            payload,
            payload_info,
            valid_payload,
            route_path,
            request,
            logger_instance=self.get_logger(route_path),
        )

        if cached_response:
            return cached_response

        r = await self.send(client_config, data=payload)
        return StreamingResponse(
            self.aiter_bytes(r, request, route_path, uid, cache_key, stream),
            status_code=r.status,
            media_type=r.headers.get("content-type"),
        )



from typing import List

from rich.console import Console, Text

RoleColor = {
    "system": "#ea4335",
    "user": "#F5A88E",
    "assistant": "#BDADFF",
    "function": "#f79393",
    "preprompt": "#f7d97f",
    "other": "#4285f4",
}

#  ----------- console tools ------------

console = Console()


def print(text="", role: str = None, end="\n", **kwargs):
    style = RoleColor.get(role, None)
    console.print(text, style=style, end=end, **kwargs)


def markdown_print(text="", role: str = None, end="\n", **kwargs):
    from rich.markdown import Markdown

    style = RoleColor.get(role, None)
    console.print(Markdown(text), style=style, end=end, **kwargs)


# ------------- parse sse -------------


def parse_sse_buffer(buffer: bytearray) -> List[str]:
    """
    Parse SSE buffer into a list of SSE message list.

    Args:
        buffer (bytearray): The SSE buffer to parse.

    Returns:
        list: A list of SSE messages.
    """
    sse_str = buffer.decode("utf-8")

    events = sse_str.split("\n\n")
    events.pop()
    return events



__version__ = "0.8.0"

from dotenv import load_dotenv
from yaml import load


def yaml_load(filepath):

    try:
        from yaml import CLoader as Loader
    except ImportError:
        from yaml import Loader
    with open(filepath, mode='r', encoding="utf-8") as stream:
        #     stream = stream.read()
        content = load(stream, Loader=Loader)
    return content


# yaml_load()
load_dotenv('.env', override=False)



import itertools
import os

import limits
from fastapi import Request

from .console import print_rate_limit_info, print_startup_info
from .content.config import setting_log
from .helper import env2dict, env2list, format_route_prefix, get_client_ip

openai_additional_start_info = {}
general_additional_start_info = {}

TIMEOUT = float(os.environ.get("TIMEOUT", "").strip() or "10")
DEFAULT_STREAM_RESPONSE = (
    os.environ.get("DEFAULT_STREAM_RESPONSE", "True").strip().lower() == "true"
)

ITER_CHUNK_TYPE = (
    os.environ.get("ITER_CHUNK_TYPE", "").strip() or "efficiency"
)  # Options: efficiency, one-by-one

CHAT_COMPLETION_ROUTE = (
    os.environ.get("CHAT_COMPLETION_ROUTE", "/v1/chat/completions").strip().lower()
)
COMPLETION_ROUTE = os.environ.get("COMPLETION_ROUTE", "/v1/completions").strip().lower()
EMBEDDING_ROUTE = os.environ.get("EMBEDDING_ROUTE", "/v1/embeddings").strip().lower()
CUSTOM_GENERAL_ROUTE = os.environ.get("CUSTOM_GENERAL_ROUTE", "").strip().lower()

CACHE_ROUTE_SET = set(env2dict("CACHE_ROUTES", []))

openai_additional_start_info['cache_routes'] = CACHE_ROUTE_SET
general_additional_start_info['cache_routes'] = CACHE_ROUTE_SET

FORWARD_CONFIG = env2dict(
    "FORWARD_CONFIG",
    [{"base_url": "https://api.openai.com", "route": "/", "type": "openai"}],
)

ENV_VAR_SEP = ","

OPENAI_BASE_URL = [
    i['base_url'] for i in FORWARD_CONFIG if i and i.get('type') == 'openai'
]
OPENAI_ROUTE_PREFIX = [
    format_route_prefix(i['route'])
    for i in FORWARD_CONFIG
    if i and i.get('type') == 'openai'
]

GENERAL_BASE_URL = [
    i['base_url'] for i in FORWARD_CONFIG if i and i.get('type') == 'general'
]
GENERAL_ROUTE_PREFIX = [
    format_route_prefix(i['route'])
    for i in FORWARD_CONFIG
    if i and i.get('type') == 'general'
]

for openai_route, general_route in zip(OPENAI_ROUTE_PREFIX, GENERAL_ROUTE_PREFIX):
    assert openai_route not in GENERAL_ROUTE_PREFIX
    assert general_route not in OPENAI_ROUTE_PREFIX

BENCHMARK_MODE = os.environ.get("BENCHMARK_MODE", "false").strip().lower() == "true"
if BENCHMARK_MODE:
    openai_additional_start_info["benchmark_mode"] = BENCHMARK_MODE

PRINT_CHAT = os.environ.get("PRINT_CHAT", "False").strip().lower() == "true"

LOG_OPENAI = os.environ.get("LOG_OPENAI", "False").strip().lower() == "true"
LOG_GENERAL = os.environ.get("LOG_GENERAL", "False").strip().lower() == "true"

CACHE_OPENAI = os.environ.get("CACHE_OPENAI", "False").strip().lower() == "true"
CACHE_GENERAL = os.environ.get("CACHE_GENERAL", "False").strip().lower() == "true"

openai_additional_start_info["LOG_OPENAI"] = LOG_OPENAI
general_additional_start_info["LOG_GENERAL"] = LOG_GENERAL


if LOG_OPENAI:
    setting_log(openai_route_prefix=OPENAI_ROUTE_PREFIX, print_chat=PRINT_CHAT)


if PRINT_CHAT:
    openai_additional_start_info["print_chat"] = True


LOG_CACHE_DB_INFO = (
    os.environ.get("LOG_CACHE_DB_INFO", "false").strip().lower() == "true"
)
CACHE_BACKEND = os.environ.get("CACHE_BACKEND", "MEMORY").strip()
CACHE_ROOT_PATH_OR_URL = os.environ.get("CACHE_ROOT_PATH_OR_URL", ".").strip()

DEFAULT_REQUEST_CACHING_VALUE = False
if CACHE_OPENAI:
    openai_additional_start_info["cache_backend"] = CACHE_BACKEND
    if not CACHE_BACKEND.lower() == 'memory':
        openai_additional_start_info["cache_root_path_or_url"] = CACHE_ROOT_PATH_OR_URL
    DEFAULT_REQUEST_CACHING_VALUE = (
        os.environ.get("DEFAULT_REQUEST_CACHING_VALUE", "false").strip().lower()
        == "true"
    )
    openai_additional_start_info[
        "default_request_caching_value"
    ] = DEFAULT_REQUEST_CACHING_VALUE

if CACHE_GENERAL:
    general_additional_start_info["cache_backend"] = CACHE_BACKEND
    if not CACHE_BACKEND.lower() == 'memory':
        general_additional_start_info["cache_root_path_or_url"] = CACHE_ROOT_PATH_OR_URL
    DEFAULT_REQUEST_CACHING_VALUE = (
        os.environ.get("DEFAULT_REQUEST_CACHING_VALUE", "false").strip().lower()
        == "true"
    )
    general_additional_start_info[
        "default_request_caching_value"
    ] = DEFAULT_REQUEST_CACHING_VALUE

IP_WHITELIST = env2list("IP_WHITELIST", sep=ENV_VAR_SEP)
IP_BLACKLIST = env2list("IP_BLACKLIST", sep=ENV_VAR_SEP)

OPENAI_API_KEY = env2dict("OPENAI_API_KEY")
FWD_KEY = env2dict("FORWARD_KEY")
LEVEL_MODELS = {int(key): value for key, value in env2dict("LEVEL_MODELS").items()}

PROXY = os.environ.get("PROXY", "").strip() or None

if PROXY:
    openai_additional_start_info["proxy"] = PROXY

GLOBAL_RATE_LIMIT = os.environ.get("GLOBAL_RATE_LIMIT", "").strip() or "inf"
RATE_LIMIT_BACKEND = os.environ.get("REQ_RATE_LIMIT_BACKEND", "").strip() or None
RATE_LIMIT_STRATEGY = (
    os.environ.get("RATE_LIMIT_STRATEGY", "fixed-window").strip() or "fixed-window"
)
req_rate_limit_dict = env2dict('REQ_RATE_LIMIT')


def get_limiter_key(request: Request):
    limiter_prefix = f"{request.scope.get('root_path')}{request.scope.get('path')}"
    fk_or_sk = request.headers.get("Authorization", "default")
    key = f"{limiter_prefix},{fk_or_sk}"
    return key


def dynamic_request_rate_limit(key: str):
    limite_prefix, fk_or_sk = key.split(',')
    key_level = FWD_KEY.get(fk_or_sk, 0)
    for route in req_rate_limit_dict:
        if key.startswith(route):
            for level_dict in req_rate_limit_dict[route]:
                if level_dict['level'] == key_level:
                    return level_dict['limit']

            break
    return GLOBAL_RATE_LIMIT


def cvt_token_rate_to_interval(token_rate_limit: str):
    if token_rate_limit:
        rate_limit_item = limits.parse(token_rate_limit)
        token_interval = (
            rate_limit_item.multiples * rate_limit_item.GRANULARITY.seconds
        ) / rate_limit_item.amount
    else:
        token_interval = 0
    return token_interval


token_rate_limit_conf = env2dict("TOKEN_RATE_LIMIT")
token_interval_conf = {}
for route, rate_limit_list in token_rate_limit_conf.items():
    token_interval_conf.setdefault(route, {})
    for level_dict in rate_limit_list:
        token_interval_conf[route][level_dict['level']] = cvt_token_rate_to_interval(
            level_dict['limit']
        )

styles = itertools.cycle(
    ["#7CD9FF", "#BDADFF", "#9EFFE3", "#f1b8e4", "#F5A88E", "#BBCA89"]
)


def show_startup():
    for base_url, route_prefix in zip(OPENAI_BASE_URL, OPENAI_ROUTE_PREFIX):
        print_startup_info(
            base_url,
            route_prefix,
            OPENAI_API_KEY,
            FWD_KEY,
            style=next(styles),
            **openai_additional_start_info,
        )
    for base_url, route_prefix in zip(GENERAL_BASE_URL, GENERAL_ROUTE_PREFIX):
        print_startup_info(
            base_url,
            route_prefix,
            "",
            "",
            style=next(styles),
            **general_additional_start_info,
        )

    print_rate_limit_info(
        RATE_LIMIT_BACKEND,
        RATE_LIMIT_STRATEGY,
        GLOBAL_RATE_LIMIT,
        req_rate_limit_dict,
        token_rate_limit_conf,
    )



from typing import List

from ..settings import (
    GENERAL_BASE_URL,
    GENERAL_ROUTE_PREFIX,
    OPENAI_BASE_URL,
    OPENAI_ROUTE_PREFIX,
    PROXY,
)
from .core import GenericForward, OpenaiForward


class ForwardManager:
    def __init__(self):
        """
        Initializes the class by creating OpenAI and Generic forward objects and their root objects.
        """
        self.openai_objs, self.openai_root_obj = self._create_forward_obj(
            OPENAI_BASE_URL, OPENAI_ROUTE_PREFIX, OpenaiForward
        )
        self.generic_objs, self.generic_root_obj = self._create_forward_obj(
            GENERAL_BASE_URL, GENERAL_ROUTE_PREFIX, GenericForward
        )
        self.root_objs = [i for i in [self.openai_root_obj, self.generic_root_obj] if i]
        if len(self.root_objs) == 2:
            raise ValueError("Only one root routing forwarding object can exist!")

    @staticmethod
    def _create_forward_obj(base_urls: List, route_prefixes: List, Forward):
        """
        Generate OpenaiForward or GenericForward objects.

        Returns:
            _objs (list): A list of Forward objects.
        """

        root_forward_obj = None
        _objs = []
        for base_url, route_prefix in zip(base_urls, route_prefixes):
            if route_prefix == "/":
                root_forward_obj = Forward(base_url, route_prefix, PROXY)
            else:
                _objs.append(Forward(base_url, route_prefix, PROXY))
        return _objs, root_forward_obj

    async def start_up(self):
        """
        Asynchronously starts up the objects by building clients for openai_objs, generic_objs, and root_objs.
        """
        [await obj.build_client() for obj in self.openai_objs]
        [await obj.build_client() for obj in self.generic_objs]
        [await obj.build_client() for obj in self.root_objs]

    async def shutdown(self):
        """
        Asynchronous shut down the client connections.
        """
        [await obj.client.close() for obj in self.openai_objs]
        [await obj.client.close() for obj in self.generic_objs]
        [await obj.client.close() for obj in self.root_objs]
